import click
from anemoi.inference.config.run import RunConfiguration
from anemoi.inference.runners.default import DefaultRunner
import earthkit.data as ekd


@click.command()
@click.option('--config', type=click.Path(exists=True), default='config.yaml', help='Inference configuration file')
def cli(config: str):
    configuration = RunConfiguration.load(config)

    ekd.config.set("cache-policy", "user")

    runner = DefaultRunner(configuration)

    import torch
    if torch.cuda.is_available():
        runner.device = "cuda"
    elif torch.backends.mps.is_available():
        runner.device = "mps"
    else:
        runner.device = "cpu"

    runner.execute()


if __name__ == '__main__':
    cli()



import click
import subprocess
import json
from bris_fiab.checkpoint.metadata import adapt_metdata

import anemoi.utils.checkpoints

description = '''
A command-line tool to update the metadata of a bris-generated checkpoint, so that it is runnable by anemoi-inference with mars input.

Note that this tool will _modify_ (not copy) the checkpoint file.
'''
default_replace_path = 'dataset.variables_metadata'


@click.command(help=description)
@click.option('--checkpoint', type=click.Path(exists=True),  required=True,
              help='Path to checkpoint to update metadata.')
@click.option('--replace-path',  type=str, default=f'{default_replace_path}',
              help=f'Dot-separated path to the key in the metadata to be replaced. Default: {default_replace_path}')
def cli(checkpoint: str, replace_path: str):

    print(f'Updating metadata in checkpoint: {checkpoint}')
    print(f'Keys updated: {replace_path}')

    metadata = anemoi.utils.checkpoints.load_metadata(checkpoint)

    adapt_metdata(metadata, replace_path)

    print(f'Loading updated metadata into {checkpoint}')

    anemoi.utils.checkpoints.replace_metadata(checkpoint, metadata)

    print(f'Updated metadata saved to checkpoint {checkpoint}')


if __name__ == "__main__":
    cli()



from dataclasses import dataclass


@dataclass
class BrisParameter:
    '''BrisParameter decodes parameters, as used in bris checkpoints'''

    parameter: str
    level: int | None = None

    @classmethod
    def from_string(cls, name: str) -> 'BrisParameter':
        elements = name.split('_')

        if len(elements) == 1:
            return BrisParameter(elements[0], None)
        elif len(elements) == 2:
            return BrisParameter(elements[0], int(elements[1]))

        raise ValueError(f"Invalid parameter name: {name}")

    def has_level(self) -> bool:
        return self.level is not None


def adapt_metdata(original_metadata: dict, replace_path: str = 'dataset.variables_metadata'):
    '''Rewrite all parameter metadata so that it is usable in a mars request. Note that this will modify the input dict!'''

    variables = original_metadata
    for p in replace_path.split('.'):
        variables = variables[p]

    for k, v in variables.items():
        if not 'mars' in v:
            continue
        p = BrisParameter.from_string(k)
        mars = v['mars']
        if p.has_level():
            mars["levtype"] = "pl"
            mars["levelist"] = p.level
            mars["param"] = p.parameter
        else:
            mars["levtype"] = "sfc"
            if p.parameter in ('z', 'lsm'):
                v["constant_in_time"] = True


if __name__ == "__main__":
    import json
    with open('inference-last.json') as f:
        metadata = json.load(f)
    print(metadata['dataset']['variables_metadata']["z"])
    adapt_metdata(metadata)
    print(metadata['dataset']['variables_metadata']["z"])



import numpy as np
import pint
from metpy.units import units
import bris_fiab.anemoi_plugins.inference.apply_adiabatic_corrections.adiabatic_correct as adiabatic_correct 


def test_correct_temperature():
    temp = pint.Quantity(np.array([10, -10, 0]), 'kelvin')

    z = pint.Quantity(np.array([0, 0, 0]), 'm^2/s^2')
    real_orog = pint.Quantity(np.array([1000, -1000, 0]), 'meter')
    diff = adiabatic_correct.get_altitude_difference(z, real_orog) # type: ignore

    # diff = pint.Quantity(np.array([1000, -1000, 0]), 'meter')

    corrected = adiabatic_correct.correct_temperature(temp, diff) # type: ignore
    expected = pint.Quantity(np.array([3.5, -3.5, 0]), 'kelvin') 

    assert np.allclose(corrected.magnitude, expected.magnitude)

def test_correct_dewpoint():
    dew = 5 * units.degC
    temp = 10 * units.degC
    corrected_temp = 3.5 * units.degC

    corrected_dew = adiabatic_correct.correct_dewpoint(dew, temp, corrected_temp) # type: ignore

    assert corrected_dew < dew
    assert corrected_dew < corrected_temp


def test_correct_surface_pressure():
    pressure = 1000 * units.hPa
    diff = -100 * units.meter

    corrected_pressure = adiabatic_correct.correct_surface_pressure(pressure, diff) # type: ignore

    assert corrected_pressure > pressure


import requests

def download(area_latlon: tuple[float, float, float, float], dest_path: str, api_key: str, dem_type: str = "SRTMGL3"):
  """
  Download topography data from OpenTopography Global DEM API.

  Args:
    area_latlon (tuple): Bounding box coordinates as (north, west, south, east).
    dest_path (str): Path to save the downloaded file.
    dem_type (str): DEM type (default: 'SRTMGL3').
  """
  url = "https://portal.opentopography.org/API/globaldem"
  params = {
    "demtype": dem_type,
    "south": area_latlon[2],
    "north": area_latlon[0],
    "west": area_latlon[1],
    "east": area_latlon[3],
    "outputFormat": "GTiff",
    "API_Key": api_key
  }
  response = requests.get(url, params=params, stream=True)
  response.raise_for_status()
  print(f"Downloading DEM of type {dem_type} for area {area_latlon}...")
  with open(dest_path, "wb") as f:
    print("Downloading...", end="", flush=True)
    count = 0
    for chunk in response.iter_content(chunk_size=8192):
      count += 1
      if count % 100 == 0:
        print(".", end="", flush=True)
      f.write(chunk)
    print(f"\nDownloaded DEM to {dest_path}")


# Adapted from code by Harrison Cook

import typing
from anemoi.inference.checkpoint import Checkpoint
import torch
import numpy as np
from copy import deepcopy
import logging
LOG = logging.getLogger(__name__)


def update(graph, model_file: str, output_file: str, model_elevation: np.ndarray | None, elevation_data: typing.Tuple[np.ndarray, np.ndarray, np.ndarray | None]):
    model = torch.load(model_file, weights_only=False,
                       map_location=torch.device('cpu'))
    # graph = torch.load(graph, weights_only=False, map_location=torch.device('cpu'))
    print(
        f"Model elevation shape: {model_elevation.shape if model_elevation is not None else None}")
    print(
        f"Elevation data shapes: {[e.shape if e is not None else None for e in elevation_data]}")

    ckpt = Checkpoint(model_file)

    supporting_arrays = ckpt._metadata._supporting_arrays

    supporting_arrays['global/cutout_mask'] = graph['data']['global/cutout_mask']
    supporting_arrays['lam_0/cutout_mask'] = np.array(
        graph['data']['lam_0/cutout_mask'])
    supporting_arrays['latitudes'] = np.array(graph['data']['latitudes'])
    supporting_arrays['longitudes'] = np.array(graph['data']['longitudes'])
    supporting_arrays['grid_indices'] = np.ones(
        graph['data']['cutout_mask'].shape, dtype=np.int64)
    supporting_arrays['lam_0/latitudes'] = elevation_data[0]
    supporting_arrays['lam_0/longitudes'] = elevation_data[1]
    if elevation_data[2] is not None:
        supporting_arrays['lam_0/correct_elevation'] = elevation_data[2]
    if model_elevation is not None:
        supporting_arrays['lam_0/model_elevation'] = model_elevation

    model = update_model(model, graph, ckpt)
    torch.save(model, output_file)

    LOG.info("Saving updated model to %s", output_file)
    from anemoi.utils.checkpoints import save_metadata

    metadata = ckpt._metadata._metadata

    metadata.dataset.data_request = {  # type: ignore
        'grid': graph['data']['global_grid'],
        'area': [90, 0.0, -90, 360],
    }

    save_metadata(
        output_file,
        metadata=metadata,
        supporting_arrays=supporting_arrays,
    )


def contains_any(key, specifications):
    contained = False
    for specification in specifications:
        if specification in key:
            contained = True
            break
    return contained


def update_state_dict(
    model, external_state_dict, keywords="", ignore_mismatched_layers=False, ignore_additional_layers=False
):
    """Update the model's stated_dict with entries from an external state_dict. Only entries whose keys contain the specified keywords are considered."""

    LOG.info("Updating model state dictionary.")

    if isinstance(keywords, str):
        keywords = [keywords]

    # select relevant part of external_state_dict
    reduced_state_dict = {
        k: v for k, v in external_state_dict.items() if contains_any(k, keywords)}
    model_state_dict = model.state_dict()

    # check layers and their shapes
    for key in list(reduced_state_dict):
        if key not in model_state_dict:
            if ignore_additional_layers:
                LOG.info(
                    "Skipping injection of %s, which is not in the model.", key)
                del reduced_state_dict[key]
            else:
                raise AssertionError(
                    f"Layer {key} not in model. Consider setting 'ignore_additional_layers = True'.")
        elif reduced_state_dict[key].shape != model_state_dict[key].shape:
            if ignore_mismatched_layers:
                LOG.info("Skipping injection of %s due to shape mismatch.", key)
                LOG.info("Model shape: %s", model_state_dict[key].shape)
                LOG.info("Provided shape: %s", reduced_state_dict[key].shape)
                del reduced_state_dict[key]
            else:
                raise AssertionError(
                    "Mismatch in shape of %s. Consider setting 'ignore_mismatched_layers = True'.", key
                )

    # update
    model.load_state_dict(reduced_state_dict, strict=False)
    return model


def update_model(model_instance, graph, checkpoint):
    # load the model from the checkpoint
    state_dict_ckpt = deepcopy(model_instance.state_dict())

    # rebuild the model with the new graph
    model_instance.graph_data = graph
    model_instance.config = checkpoint._metadata._config
    model_instance._build_model()

    # reinstate the weights, biases and normalizer from the checkpoint
    # reinstating the normalizer is necessary for checkpoints that were created
    # using transfer learning, where the statistics as stored in the checkpoint
    # do not match the statistics used to build the normalizer in the checkpoint.
    model_instance = update_state_dict(
        model_instance, state_dict_ckpt, keywords=[
            "bias", "weight", "processors.normalizer"]
    )

    return model_instance



import os
import pint
from metpy.units import units
import earthkit.data as ekd
import metpy.calc
import numpy as np

from bris_fiab.anemoi_plugins.inference.apply_adiabatic_corrections.apply_adiabatic_corrections import AdiabaticCorrector

class AbstractAdiabaticCorrectorTest:
    grib_file = 'none.grib'  # Override in subclasses

    def setup_method(self):
        test_file = os.path.join(os.path.dirname(__file__), 'test_data', self.grib_file)
        self.test_data = ekd.from_source('file', test_file)
        z = self.test_data.sel(param='z')[0]

        model_elevation = np.zeros(z.shape) * units.meter
        self.correct_elevation = np.full(z.shape, 1000) * units.meter

        corrector = AdiabaticCorrector(
            model_elevation=model_elevation, correct_elevation=self.correct_elevation)

        self.corrected = corrector.apply(self.test_data)  # type: ignore
    

class TestAdiabaticCorrectorVariableTimes(AbstractAdiabaticCorrectorTest):
    grib_file = '1.grib'
 
    def test_same_output_size(self):
        assert len(self.corrected) == len(self.test_data)  # type: ignore

    def test_2t(self):
        original_t = self.test_data.sel(param='2t', levtype='sfc').to_numpy()
        corrected_t = self.corrected.sel(param='2t', levtype='sfc').to_numpy()
        np.allclose(corrected_t, original_t - 6.5)

    def test_2d(self):
        original_d = self.test_data.sel(param='2d', levtype='sfc').to_numpy()
        corrected_d = self.corrected.sel(param='2d', levtype='sfc').to_numpy()
        assert np.all(corrected_d < original_d)
        corrected_t = self.corrected.sel(param='2t', levtype='sfc').to_numpy()
        assert np.all(corrected_d < corrected_t)

    def test_sp(self):
        original_sp = self.test_data.sel(param='sp', levtype='sfc').to_numpy()
        corrected_sp = self.corrected.sel(param='sp', levtype='sfc').to_numpy()
        assert np.all(corrected_sp < original_sp)

    def test_others(self):
        for n, d in enumerate(self.test_data):
            if d.metadata('param') in ['2t', '2d', 'sp']:
                # these are supposed to change
                continue
            original = d.to_numpy()
            corrected = self.corrected[n].to_numpy()  # type: ignore
            np.allclose(corrected, original)


class TestAdiabaticCorrectorNonChangingTimes(AbstractAdiabaticCorrectorTest):
    grib_file = '2.grib'

    def test_z(self):
        corrected_z = self.corrected.sel(param='z', levtype='sfc').to_numpy()
        expected_z = metpy.calc.height_to_geopotential(self.correct_elevation).magnitude
        assert np.allclose(corrected_z, expected_z)



from typing import Tuple
import numpy as np
# from scipy.spatial import cKDTree
from metpy.interpolate import interpolate_to_points


def _sph2cart(lat_deg: np.ndarray, lon_deg: np.ndarray) -> np.ndarray:
    """Convert spherical coordinates (latitude and longitude in degrees) to Cartesian coordinates (x, y, z)."""
    lat = np.deg2rad(lat_deg)
    lon = np.deg2rad(lon_deg)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)
    return np.stack([x, y, z], axis=-1)


def interpolate_to_grid(src_latitude: np.ndarray, src_longitude: np.ndarray, src_vals: np.ndarray,
                        dst_latitude: np.ndarray, dst_longitude: np.ndarray, interp_type='nearest') -> np.ndarray:
    """
    Interpolate values from a source grid to a destination grid using specified interpolation method.
    Parameters:
      src_latitude (np.ndarray): 2D Array of latitudes for the source points. Shape should match src_vals.
      src_longitude (np.ndarray): 2D Array of longitudes for the source points. Shape should match src_latitude and src_vals.
      src_vals (np.ndarray): 2D Array of values at the source points. Shape should match src_latitude and src_longitude.
      dst_latitude (np.ndarray): 2D Array of latitudes for the destination grid. Shape should match dst_longitude.
      dst_longitude (np.ndarray): 2D Array of longitudes for the destination grid. Shape should match dst_latitude.
      interp_type (str, optional): Interpolation method to use. Default is 'nearest'. Other methods may be supported depending on implementation.
                see metpy.interpolate.interpolate_to_points for options.
    Returns:
      np.ndarray: Interpolated values at the destination grid points, with shape matching dst_latitude.
    """
    src_pts = _sph2cart(src_latitude.ravel(),
                        src_longitude.ravel())
    dst_pts = _sph2cart(dst_latitude.ravel(),
                        dst_longitude.ravel())
    dst_vals = interpolate_to_points(src_pts, src_vals.ravel(), dst_pts,
                                     interp_type=interp_type)
    dst_vals = dst_vals.reshape(dst_latitude.shape)
    return dst_vals



# Adapted from code by Harrison Cook
#
# This script builds a stretched graph from latitude and longitude data.


edge_attrs = {
    "edge_length": {
      "_target_": "anemoi.graphs.edges.attributes.EdgeLength",
      "norm": "unit-max"
    },
    "edge_dirs": {
      "_target_": "anemoi.graphs.edges.attributes.EdgeDirection",
      "norm": "unit-std"
    }
}


def combine_nodes(latitudes, longitudes, global_lats, global_lons):
    from anemoi.datasets.grids import cutout_mask
    import torch
    import numpy as np

    
    _mask = cutout_mask(latitudes, longitudes, global_lats, global_lons)
    lats = np.concatenate([latitudes, global_lats[_mask]])
    lons = np.concatenate([longitudes, global_lons[_mask]])
    mask = torch.tensor([True] * len(latitudes) + [False] * sum(_mask), dtype=torch.bool)
    return lats, lons, mask, _mask


def build_stretched_graph(latitudes, longitudes, global_grid: str, lam_resolution: int, global_resolution: int, margin_radius_km: int):

    from torch_geometric.data import HeteroData
    from anemoi.graphs.nodes import LatLonNodes, StretchedTriNodes
    from anemoi.graphs.edges import KNNEdges, MultiScaleEdges
    from anemoi.utils.grids import grids

    import torch
    import numpy as np

    assert latitudes.ndim == 1
    assert longitudes.ndim == 1
    assert len(latitudes) == len(longitudes)
    
    global_points = grids(global_grid)
    lats, lons, mask, _mask = combine_nodes(latitudes, longitudes, global_points["latitudes"], global_points["longitudes"])

    graph = LatLonNodes(lats, lons, name="data").update_graph(HeteroData())
    graph["data"]["global_grid"] = global_grid
    graph["data"]["cutout_mask"] = mask
    graph["data"]["latitudes"] = lats
    graph["data"]["longitudes"] = lons
    graph["data"]["global/cutout_mask"] = _mask
    graph["data"]["lam_0/cutout_mask"] =  torch.tensor([True] * len(latitudes))

    # All of the following can easily be moved to a configuration file and substituted by:
    # graph = GraphCreator("recipe_forecast_in_a_box.yaml").update_graph(graph)
    hidden = StretchedTriNodes(
        lam_resolution=lam_resolution, 
        global_resolution=global_resolution, 
        margin_radius_km=margin_radius_km,
        reference_node_name="data", 
        mask_attr_name="cutout_mask", 
        name="hidden",
    )
    enc = KNNEdges("data", "hidden", num_nearest_neighbours=12)
    proc = MultiScaleEdges("hidden", "hidden", x_hops=1, scale_resolutions=list(range(1, lam_resolution + 1)))
    dec = KNNEdges("hidden", "data", num_nearest_neighbours=1)

    graph = hidden.update_graph(graph)

    graph = enc.update_graph(graph, edge_attrs)
    graph = proc.update_graph(graph, edge_attrs)
    graph = dec.update_graph(graph, edge_attrs)

    return graph




# import argparse
# def parse_args():
#     parser = argparse.ArgumentParser(description="Create a graph from latitudes and longitudes.")
#     parser.add_argument("lats", type=str, help="Path to the lats.")
#     parser.add_argument("lons", type=str, help="Path to the lons.")
#     parser.add_argument("--global_grid", type=str, default="n320", help="Global grid resolution (default: n320).")
#     parser.add_argument("--lam_resolution", type=int, default=8, help="LAM resolution (default: 8).")
#     parser.add_argument("--output", type=str, default="graph.pt", help="Path to the output graph file.")

#     return parser.parse_args()

# if __name__ == "__main__":
#     args = parse_args()
#     import numpy as np
#     import torch

#     lat = np.load(args.lats)
#     lon = np.load(args.lons)

#     graph = build_stretched_graph(lat, lon, global_grid=args.global_grid, lam_resolution=args.lam_resolution)
#     torch.save(graph, args.output)

#     torch.load(args.output, weights_only=False, map_location=torch.device('cpu'))


from anemoi.inference.inputs.mars import MarsInput
from anemoi.inference.types import Date
from anemoi.inference.context import Context
from anemoi.inference.processor import Processor
from earthkit.data.sources.array_list import ArrayField
import earthkit.data as ekd
import rioxarray
import rasterio
import numpy as np
import typing
import zipfile
import scipy.interpolate
from scipy.spatial import Delaunay
from dataclasses import dataclass
from bris_fiab.checkpoint.interpolate import interpolate_to_grid


@dataclass
class Topography:
    '''A simple holder for output topography data'''
    x_values: np.ndarray  # longitudes in a two-dimensional array
    y_values: np.ndarray  # latitudes in a two-dimensional array
    elevation: np.ndarray | None  # elevation in a two-dimensional array
    spatial_ref: typing.Any  # rasterio.crs.CRS

    @classmethod
    def from_topography_file(cls, topography_file: str | rasterio.MemoryFile):
        topography = rioxarray.open_rasterio(topography_file)

        x_values, y_values = make_two_dimensional(
            topography['x'].values,  # type: ignore
            topography['y'].values,  # type: ignore
        )

        if x_values.shape != y_values.shape or y_values.shape != topography.values[0].shape:
            raise ValueError(
                "topography x, y, and elevation must have the same shape")

        return Topography(
            x_values=x_values,
            y_values=y_values,
            elevation=topography.values[0],  # type: ignore
            spatial_ref=topography.spatial_ref,  # type: ignore
        )

    @classmethod
    def from_topography_file_to_grid(cls, topography_file: str | rasterio.MemoryFile, latitudes: np.ndarray, longitudes: np.ndarray) -> 'Topography':
        topo = cls.from_topography_file(topography_file)

        values = interpolate_to_grid(topo.y_values, topo.x_values,
                                     topo.elevation, latitudes, longitudes)
        return Topography(
            x_values=longitudes,
            y_values=latitudes,
            elevation=values,
            spatial_ref=topo.spatial_ref,  # This is not very useful without the original grid
        )

    @classmethod
    def from_supporting_array(cls, context: Context) -> 'Topography':
        """Create a Topography instance from a supporting array in the checkpoint."""
        latitudes = context.checkpoint.supporting_arrays['lam_0/latitudes']
        longitudes = context.checkpoint.supporting_arrays['lam_0/longitudes']
        try:
            elevation = context.checkpoint.supporting_arrays['lam_0/correct_elevation']
        except KeyError:
            elevation = None

        return Topography(
            x_values=longitudes,  # type: ignore
            y_values=latitudes,  # type: ignore
            elevation=elevation,  # type: ignore
            spatial_ref=None,  # type: ignore
        )


def downscaler(ix: np.ndarray, iy: np.ndarray, ox: np.ndarray, oy: np.ndarray) -> typing.Callable[[np.ndarray], np.ndarray]:

    if len(ix.shape) == 1:
        if len(iy.shape) != 1:
            raise ValueError("ix must be 1D if iy is 1D")
        ix, iy = make_two_dimensional(ix, iy)
    ipoints = np.column_stack((iy.flatten(), ix.flatten()))

    triangulation = Delaunay(ipoints)

    if len(ox.shape) == 1:
        if len(oy.shape) != 1:
            raise ValueError("ox must be 1D if oy is 1D")
        ox, oy = make_two_dimensional(ox, oy)
    opoints = np.column_stack((oy.flatten(), ox.flatten()))

    def interpolate(values: np.ndarray) -> np.ndarray:
        interpolator = scipy.interpolate.LinearNDInterpolator(
            triangulation, values.flatten())
        return interpolator(opoints).reshape(ox.shape)
    return interpolate


class DownscalePreProcessor(Processor):
    def __init__(self, context: Context, **kwargs):
        if 'orography_file' in kwargs:
            self._topography = Topography.from_topography_file(
                kwargs['orography_file'])
        else:
            self._topography = Topography.from_supporting_array(context)

        super().__init__(context, **kwargs)

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:  # type: ignore
        return downscale(fields, self._topography.x_values, self._topography.y_values)


class DownscaledMarsInput(MarsInput):
    def __init__(self, context: Context, **kwargs):
        """Initialize the Downscaled Mars Input.

        Parameters
        ----------
        context : Context
            The context in which the input operates.
        mars_options : dict, optional
            Options for MARS retrieval. Keys will be converted to strings.
        """
        if "grid" in kwargs:
            grid = kwargs["grid"]
            if isinstance(grid, str):
                if len(grid) > 0 and grid[0] in ("O", "N", "H"):
                    raise ValueError(
                        "only regular grids are supported for downscaling")

        if 'orography_file' in kwargs:
            self._topography = Topography.from_topography_file(
                kwargs['orography_file'])
            del kwargs['orography_file']
        else:
            self._topography = Topography.from_supporting_array(context)

        super().__init__(context, **kwargs)

    def retrieve(
        self, variables: typing.List[str], dates: typing.List[Date]
    ) -> typing.Any:
        original: ekd.FieldList = super().retrieve(variables, dates)  # type: ignore
        return downscale(original, self._topography.x_values, self._topography.y_values)


def downscale(source_ds: ekd.FieldList, output_x_values: np.ndarray, output_y_values: np.ndarray) -> ekd.FieldList:
    fields = []

    field: ekd.FieldList = source_ds.sel(param="z")  # type: ignore
    latlon = field.to_latlon()

    downscale = downscaler(
        iy=latlon['lat'],  # type: ignore
        ix=latlon['lon'],  # type: ignore
        oy=output_y_values,
        ox=output_x_values,
    )

    metadata_overrides = {
        'Ni': output_x_values.shape[1],
        'Nj': output_y_values.shape[0],
        'latitudeOfFirstGridPointInDegrees': output_y_values[0, 0],
        'latitudeOfLastGridPointInDegrees': output_y_values[-1, 0],
        'longitudeOfFirstGridPointInDegrees': output_x_values[0, 0],
        'longitudeOfLastGridPointInDegrees': output_x_values[0, -1],
    }

    for field in source_ds:  # type: ignore
        # name: str = field.metadata("shortName")  # type: ignore

        # original_data = source_ds.sel(param=name).to_numpy()  # type: ignore
        original_data = field.to_numpy()

        data = downscale(original_data)  # type: ignore
        metadata = field.metadata().override(**metadata_overrides)  # type: ignore
        af = ArrayField(data, metadata)
        fields.append(af)

    # return source_ds
    ret = ekd.FieldList.from_fields(fields)
    return ret


def make_two_dimensional(x_values: np.ndarray, y_values: np.ndarray) -> typing.Tuple[np.ndarray, np.ndarray]:
    x = np.tile(x_values, (len(y_values), 1))
    y = np.transpose(np.tile(y_values, (len(x_values), 1)))
    return x, y


def geopotential_to_height(z):
    earth_radius = 6371008.7714
    g = 9.80665
    return (z * earth_radius) / (g * earth_radius - z)


def height_to_geopotential(h):
    earth_radius = 6371008.7714
    g = 9.80665
    return (g * earth_radius * h) / (earth_radius + h)



from metpy.units import units
import metpy.calc
import pint


def get_altitude_difference(model_geopotential: pint.Quantity, real_orography: pint.Quantity) -> pint.Quantity:
    model_orography = metpy.calc.geopotential_to_height(model_geopotential)
    return (real_orography - model_orography)

def convert_to_geopotential(height: pint.Quantity) -> pint.Quantity:
    return metpy.calc.height_to_geopotential(height)

def correct_temperature(temperature: pint.Quantity, height_difference: pint.Quantity) -> pint.Quantity:
    factor = 0.0065 * units.kelvin / units.meter
    correct_temperature = temperature - (factor * height_difference)
    return correct_temperature.to(temperature.units)

def correct_dewpoint(original_dewpoint: pint.Quantity, original_temperature: pint.Quantity, corrected_temperature: pint.Quantity) -> pint.Quantity:
    relhum = metpy.calc.relative_humidity_from_dewpoint(original_temperature, original_dewpoint)
    corrected_dewpoint = metpy.calc.dewpoint_from_relative_humidity(corrected_temperature, relhum)
    return corrected_dewpoint.to(original_dewpoint.units)

def correct_surface_pressure(original_pressure: pint.Quantity, height_difference: pint.Quantity) -> pint.Quantity:
    return metpy.calc.add_height_to_pressure(original_pressure, height_difference).to(original_pressure.units)



import sys
import os
import click
from bris_fiab.checkpoint import graph


@click.command()
@click.option('--topography-file', type=click.Path(exists=True), default=None)
@click.option('--area-latlon', type=(float, float, float, float, float), default=None, help='Area defined by (north, west, south, east, resolution)')
@click.option('--original-checkpoint', type=click.Path(exists=True))
@click.option('--create-checkpoint', type=click.Path())
@click.option('--without-model-elevation', is_flag=True, default=False, help='Do not download and add model elevation data to the checkpoint.')
@click.option('--save-graph-to', type=click.Path(), default='', help='Save the graph file to this path in addition to integrating it into the checkpoint.')
@click.option('--save-latlon', type=bool, default=False, help='Whether to save the latitude/longitude to files.')
@click.option('--lam-resolution', type=int, default=10)
@click.option('--global-resolution', type=int, default=7)
@click.option('--margin-radius-km', type=int, default=11)
def cli(topography_file: str | None, area_latlon: tuple[float, float, float, float, float] | None, original_checkpoint: str, create_checkpoint: str, without_model_elevation: bool, save_graph_to: str, save_latlon: bool, lam_resolution: int, global_resolution: int, margin_radius_km: int):

    for f in (topography_file, original_checkpoint):
        if f is not None:
            if not os.path.exists(f):
                print(f'File {f} does not exist.')
                sys.exit(1)

    if topography_file is None and area_latlon is None:
        print(
            'Either topography_file or area_latlon must be provided.')
        sys.exit(1)

    graph.run(
        topography_file=topography_file,
        original_checkpoint=original_checkpoint,
        new_checkpoint=create_checkpoint,
        add_model_elevation=not without_model_elevation,
        save_graph_to=save_graph_to,
        save_latlon=save_latlon,
        graph_config=graph.GraphConfig(
            lam_resolution=lam_resolution,
            global_resolution=global_resolution,
            margin_radius_km=margin_radius_km,
            area_latlon=area_latlon
        )
    )


if __name__ == "__main__":
    cli()



from anemoi.inference.processor import Processor
from anemoi.inference.context import Context
from metpy.units import units
import earthkit.data as ekd
import pint
import bris_fiab.anemoi_plugins.inference.apply_adiabatic_corrections.adiabatic_correct as adiabatic_correct


class AdiabaticCorrectionPreProcessor(Processor):
    def __init__(self, context: Context, **kwargs):
        model_elevation = context.checkpoint.supporting_arrays['lam_0/model_elevation']
        correct_elevation = context.checkpoint.supporting_arrays['lam_0/correct_elevation']

        self._corrector = AdiabaticCorrector(
            model_elevation * units.meters,
            correct_elevation * units.meters
        )
        super().__init__(context, **kwargs)

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:
        return self._corrector.apply(fields)


class AdiabaticCorrector:
    def __init__(self, model_elevation: pint.Quantity, correct_elevation: pint.Quantity):
        self._correct_elevation = correct_elevation
        self._altitude_difference = correct_elevation - model_elevation

    def apply(self, fields: ekd.FieldList) -> ekd.FieldList:

        corrected_temperatures = {}
        original_temperatures = {}
        temperatures = fields.sel(param='2t', levtype='sfc')
        for t in temperatures:
            values = t.to_numpy() * units.kelvin
            key = t.datetime()['valid_time']
            original_temperatures[key] = values
            corrected_temperature = adiabatic_correct.correct_temperature(values, self._altitude_difference)
            corrected_temperatures[key] = corrected_temperature
        
        ret = []
        for field in fields:
            # Only correct surface fields
            if field.metadata('levtype') != 'sfc':
                ret.append(field)
                continue
            key = field.datetime()['valid_time']
            param = field.metadata('param')
            if param == '2t':
                values = corrected_temperatures[key]
                ret.append(field.copy(values=values.magnitude))  # type: ignore
            elif param == '2d':
                old_values = field.to_numpy() * units.kelvin
                values = adiabatic_correct.correct_dewpoint(
                    original_dewpoint=old_values, 
                    original_temperature=original_temperatures[key], 
                    corrected_temperature=corrected_temperatures[key]
                )
                ret.append(field.copy(values=values.magnitude))  # type: ignore
            elif param == 'sp':
                old_values = pint.Quantity(field.to_numpy(), field.metadata('units'))
                values = adiabatic_correct.correct_surface_pressure(old_values, self._altitude_difference)  # type: ignore
                ret.append(field.copy(values=values.magnitude))  # type: ignore
            elif param == 'z':
                values =  adiabatic_correct.convert_to_geopotential(self._correct_elevation)
                ret.append(field.copy(values=values.magnitude))  # type: ignore
            else:
                ret.append(field)

        return ekd.SimpleFieldList(ret)



import json
import click
import rioxarray
import xarray as xr
import numpy as np
from typing import Dict
import pydantic


class VariableConfig(pydantic.BaseModel):
    variable_name: str
    attributes: Dict[str, object]


class SurfaceVariablesConfig(pydantic.BaseModel):
    variables: Dict[str, VariableConfig]

class PressureLevelVariablesConfig(pydantic.BaseModel):
    levels: list[int]
    variables: Dict[str, VariableConfig]

class VariablesConfig(pydantic.BaseModel):
    sfc: SurfaceVariablesConfig
    pl: PressureLevelVariablesConfig

class MkGridConfig(pydantic.BaseModel):
    variables: VariablesConfig


@click.command()
@click.option('--grid', type=click.Path(exists=True), help='Grid to convert to')
@click.option('--config', type=click.Path(exists=True), default='etc/mkgrid.json', help='Configuration file for variable mapping')
@click.argument('input', type=click.Path(exists=True))
@click.argument('output', type=click.Path())
def cli(grid: str, config: str, input: str, output: str):
    with open(config) as f:
        config_json = json.load(f)
        met_variables = MkGridConfig.model_validate(config_json)

    data = xr.open_dataset(input)

    if grid:
        elevation = rioxarray.open_rasterio(grid)
        x: np.ndarray = elevation.x.values # type: ignore
        y: np.ndarray = elevation.y.values # type: ignore
        spatial_ref = elevation['spatial_ref'] # type: ignore
    else:
        x = np.unique(data.longitude.values)
        y = np.unique(data.latitude.values)[::-1]
        spatial_ref = xr.DataArray(
            data=0,
            attrs={
                'crs_wkt': 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]', 
                'semi_major_axis': 6378137.0, 
                'semi_minor_axis': 6356752.314245179, 
                'inverse_flattening': 298.257223563, 
                'reference_ellipsoid_name': 'WGS 84', 
                'longitude_of_prime_meridian': 0.0, 
                'prime_meridian_name': 'Greenwich', 
                'geographic_crs_name': 'WGS 84', 
                'horizontal_datum_name': 'World Geodetic System 1984', 
                'grid_mapping_name': 'latitude_longitude', 
                'spatial_ref': 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]', 
                # 'GeoTransform': f'29.999583333285614 0.025 0.0 -7.9995833333178865 0.0 -0.025'
                'GeoTransform': f'{x[0]} {(x[1] - x[0]):.3g} 0.0 {y[-1]} 0.0 {(y[-1] - y[-2]):.3g}'
            }
        )

    size = len(x) * len(y)
    time_count = len(data['time'])

    variables = {
        'spatial_ref': spatial_ref, # type: ignore
        'forecast_reference_time': xr.DataArray(
            np.datetime64(data['time'].values[0]),
            dims=(),
            attrs={
                'long_name': 'forecast reference time',
                'standard_name': 'forecast_reference_time',
            }
        )
    }

    coords = {
        'time': xr.DataArray(
            data['time'].values,
            dims='time',
            attrs={
                'standard_name': 'time'
            }
        ),
        'lat': xr.DataArray(
            y,
            dims='lat',
            attrs={
                'units': 'degree',
                'standard_name': 'latitude'
            }
        ),
        'lon': xr.DataArray(
            x,
            dims='lon',
            attrs={
                'units': 'degree',
                'standard_name': 'longitude'
            }
        ),
        'pl': xr.DataArray(
            met_variables.variables.pl.levels,
            dims='pl',
            attrs={
                'units': 'hPa',
                'standard_name': 'air_pressure',
                'long_name': 'pressure level',
            }
        )
    }

    for variable, cfg in met_variables.variables.sfc.variables.items():
        if variable not in data.data_vars:
            print(f"Variable {variable} not found in input data.")
            continue
        if not cfg.variable_name:
            # print(f"Variable {variable} is not configured.")
            continue

        param_data = data[variable].values[:, :size].reshape(
            (time_count, len(y), len(x)))
        param = xr.DataArray(
            param_data,
            coords=[data['time'], y, x],
            dims=['time', 'lat', 'lon'],
            attrs={**cfg.attributes, "grid_mapping": "spatial_ref"}
        )
        variables[cfg.variable_name] = param

    for variable, cfg in met_variables.variables.pl.variables.items():
        if not cfg.variable_name:
            print(f"Variable {variable} is not configured.")
            continue

        variable_names = [f'{variable}_{level}' for level in met_variables.variables.pl.levels]        
        param_data = [data[vn].values[:, :size].reshape((time_count, len(y), len(x))) for vn in variable_names]
        param_data = np.stack(param_data, axis=1)

        param = xr.DataArray(
            param_data,
            coords=[data['time'], met_variables.variables.pl.levels, y, x],
            dims=['time', 'pl', 'lat', 'lon'],
            attrs={**cfg.attributes, "grid_mapping": "spatial_ref"}
        )
        variables[cfg.variable_name] = param


    ds = xr.Dataset(
        variables,
        coords=coords,
    )

    ds.to_netcdf(output)
    print(ds)


if __name__ == "__main__":
    cli()



import click
import json
from bris_fiab.orography import download
import os

DEFAULT_API_KEY_FILE = '.opentopographyrc'

def find_api_key_file(filename: str=DEFAULT_API_KEY_FILE):
  # Check current working directory
  cwd_path = os.path.join(os.getcwd(), filename)
  if os.path.exists(cwd_path):
    return cwd_path
  # Check home directory
  home_path = os.path.join(os.path.expanduser('~'), filename)
  if os.path.exists(home_path):
    return home_path
  return None

def read_api_key(filepath: str) -> str | None:
  with open(filepath, 'r') as f:
    data = json.load(f)
  return data.get('api_key')

@click.command(
    help=(
      "Download a DEM from OpenTopography (https://opentopography.org).\n\n"
      "The API key file must be a JSON file with the following format:\n"
      '{\n  "api_key": "YOUR_API_KEY_HERE"\n}\n\n'
      "Default api_key file is '.opentopographyrc' in the current or home directory.\n\n"
      "Create an account and get an API key from https://portal.opentopography.org/login.\n"
    )
  )
@click.option('--area-latlon', type=(float, float, float, float), required=True, help='Bounding box coordinates as (north, west, south, east)')
@click.option('--api-key-file', type=click.Path(exists=True), default=None, show_default=False, help='Path to JSON file containing the API key')
@click.option('--dem-type', type=click.Choice(['SRTMGL3', 'SRTMGL1', 'AW3D30', 'TDM1', 'COP30']), default='SRTMGL3', show_default=True, help='Type of DEM to download')
@click.option('--dest-path', type=click.Path(), default='orography.tif', show_default=True, help='Destination path for the downloaded DEM file')
def main(area_latlon: tuple[float, float, float, float], api_key_file: str | None, dem_type: str, dest_path: str):
  if api_key_file is None:
    api_key_file = find_api_key_file()
    if not api_key_file:
      print("API key file not found in current or home directory.")
      return
    print(f"Using API key file: {api_key_file}")

  api_key = read_api_key(api_key_file)
  if not api_key:
    print("API key not found in the provided file.")
    return

  print(f"North: {area_latlon[0]}")
  print(f"West: {area_latlon[1]}")
  print(f"South: {area_latlon[2]}")
  print(f"East: {area_latlon[3]}")
  print(f"Using API key from: {api_key_file}")


  download.download(area_latlon, dest_path, api_key, dem_type)

if __name__ == "__main__":
  main()



import numpy as np
from bris_fiab.anemoi_plugins.inference.downscale.downscale import Topography, make_two_dimensional
import earthkit.data as ekd
from dataclasses import dataclass
from .make_graph import build_stretched_graph
from .update import update
from .elevation import get_model_elevation, get_model_elevation_mars_grid


@dataclass
class GraphConfig:
    lam_resolution: int = 10
    global_resolution: int = 7
    margin_radius_km: int = 11
    area_latlon: tuple[float, float, float, float, float] | None = None


def run(topography_file: str | None, original_checkpoint: str, new_checkpoint: str, add_model_elevation: bool, save_graph_to: str, save_latlon: bool, graph_config: GraphConfig = GraphConfig()):
    elevation: np.ndarray | None = None

    if graph_config.area_latlon is not None:
        lat, lon = _get_lat_lon_from_area(graph_config.area_latlon)
        elevation = None
    elif topography_file is not None:
        lat, lon, elevation = _get_latlon(topography_file)
    else:
        raise ValueError(
            'Either topography_file or area_latlon must be provided.')

    if save_latlon:
        with open('latitudes.npy', 'wb') as f:
            np.save(f, lat)
        with open('longitudes.npy', 'wb') as f:
            np.save(f, lon)
        if elevation is not None:
            with open('elevations.npy', 'wb') as f:
                np.save(f, elevation)

    if elevation is None and topography_file is not None:
        print(
            f'Loading correct elevation from {topography_file} interpolated to the mars grid')
        _tlat, _tlon, elevation = _get_topography_on_grid(
            topography_file, lat, lon)

    if elevation is None:
        print(f'No elevation data found in {topography_file}')

    if add_model_elevation:
        # model_elevation = get_model_elevation(lat, lon)
        if graph_config.area_latlon is not None:
            model_elevation = get_model_elevation_mars_grid(
                graph_config.area_latlon)
        else:
            model_elevation = get_model_elevation(lat, lon)
    else:
        model_elevation = None

    graph = build_stretched_graph(
        lat.flatten(), lon.flatten(),
        global_grid='n320',
        lam_resolution=graph_config.lam_resolution,
        global_resolution=graph_config.global_resolution,
        margin_radius_km=graph_config.margin_radius_km
    )

    if save_graph_to:
        import torch
        torch.save(graph, save_graph_to)
        print('saved graph')
    # graph = torch.load(args.output, weights_only=False, map_location=torch.device('cpu'))

    update(graph, original_checkpoint, new_checkpoint,
           model_elevation, (lat, lon, elevation))


def _get_latlon(topography_file: str) -> tuple[np.ndarray, np.ndarray, np.ndarray | None]:
    topo = Topography.from_topography_file(topography_file)
    return topo.y_values, topo.x_values, topo.elevation


def _get_topography_on_grid(topography_file: str, latitude: np.ndarray, longitude: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray | None]:
    topo = Topography.from_topography_file_to_grid(
        topography_file, latitude, longitude)
    return topo.y_values, topo.x_values, topo.elevation


def _get_lat_lon_from_area(area_latlon: tuple[float, float, float, float, float]) -> tuple[np.ndarray, np.ndarray]:
    """The function use earthkit.data to download lat/lon for the specified area from Mars.
    area_latlon: (north, west, south, east, resolution)
    returns: lat, lon
    """

    # resolution is area_latlon[4]
    # return lat, lon arrays
    area = [area_latlon[0], area_latlon[1], area_latlon[2], area_latlon[3]]
    ds = ekd.from_source(  # type: ignore
        'mars',
        {
            'AREA': area,
            'GRID': f"{area_latlon[4]}/{area_latlon[4]}",
            'param': '2t',
            'date': -34,
            'stream': 'oper',
            'type': 'an',
            'levtype': 'sfc',
        }
    )

    lat, lon = ds[0].grid_points()  # type: ignore
    data = ds[0].to_numpy()  # type: ignore
    print(
        f"get_lat_lon_from_area: Downloaded lat/lon with shapes {lat.shape}, {lon.shape}, data shape {data.shape}")

    return lat.reshape(data.shape), lon.reshape(data.shape)  # type: ignore



import numpy as np
import earthkit.data as ekd
import metpy.calc
from metpy.units import units
from bris_fiab.anemoi_plugins.inference.downscale.downscale import Topography, downscaler


def get_model_elevation(lats: np.ndarray, lons: np.ndarray) -> np.ndarray:
    '''Fetch model elevation from ECMWF MARS, interpolated to the given latitudes and longitudes.'''
    max_lat = np.ceil(np.max(lats) * 10) / 10
    min_lat = np.floor(np.min(lats) * 10) / 10
    max_lon = np.ceil(np.max(lons) * 10) / 10
    min_lon = np.floor(np.min(lons) * 10) / 10
    request = {
        'area': [max_lat, min_lon, min_lat, max_lon],
        'date': -10,
        'expver': '0001',
        'grid': '0.05/0.05',
        'levtype': 'sfc',
        'param': ['z'],
        'step': 0,
        'time': '0000'
    }

    raw_data = ekd.from_source('mars', request)[0]

    geopotential = units.Quantity(raw_data.to_numpy(), 'm^2/s^2')
    print(f'Geopotential min/max: {geopotential.min()}/{geopotential.max()}')
    height = metpy.calc.geopotential_to_height(geopotential)

    latlons = raw_data.to_latlon()
    return downscaler(latlons['lon'], latlons['lat'], lons, lats)(height).astype('int16')


def get_model_elevation_mars_grid(area_latlon: tuple[float, float, float, float, float]) -> np.ndarray:
    """The function use earthkit.data to download model elevation for the specified area from Mars.
    area_latlon: (north, west, south, east, resolution)
    returns: elevation
    """

    # resolution is area_latlon[4]
    # return lat, lon arrays
    area = [area_latlon[0], area_latlon[1], area_latlon[2], area_latlon[3]]
    ds = ekd.from_source(  # type: ignore
        'mars',
        {
            'AREA': area,
            'GRID': f"{area_latlon[4]}/{area_latlon[4]}",
            'param': 'z',
            'date': -34,
            'stream': 'oper',
            'type': 'an',
            'levtype': 'sfc',
        }
    )
    xarr = ds.to_xarray()  # type: ignore
    geopotential = units.Quantity(ds[0].to_numpy(), 'm^2/s^2')
    height = metpy.calc.geopotential_to_height(geopotential)
    return height.astype('int16')  # type: ignore
